/*
 * tnef.c -- extract files from microsoft TNEF format
 *
 * Copyright (C)1999, 2000, 2001 Mark Simpson <damned@world.std.com>
 * Copyright (C)1998 Thomas Boll  <tb@boll.ch>		[ORIGINAL AUTHOR]
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you can either send email to this
 * program's maintainer or write to: The Free Software Foundation,
 * Inc.; 59 Temple Place, Suite 330; Boston, MA 02111-1307, USA.
 *
 * Commentary:
 *       scans tnef file and extracts all attachments
 *       attachments are written to their original file-names if possible
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>

#if STDC_HEADERS
#  include <string.h>
#else
#  if HAVE_MEMMOVE
#    define memmove(d, s, n) bcopy ((s), (d), (n))
#  endif/* HAVE_MEMMOVE */
#endif /* STDC_HEADERS */

#include <stdarg.h>

#include "strdup.h"

#include "tnef.h"

#include "alloc.h"
#include "path.h"

/* To quiet compiler define tempnam */
extern char*
tempnam(const char*, const char*);

/* Needed to transform char buffers into little endian numbers */
#define GETINT32(p)    (uint32)((uint8)(p)[0]           \
                                +((uint8)(p)[1]<<8)     \
                                +((uint8)(p)[2]<<16)    \
                                +((uint8)(p)[3]<<24))
#define GETINT16(p)    (uint16)((uint8)(p)[0]+((uint8)(p)[1]<<8))

/* Global variables, used by all (or nearly all) functions */
static int8 g_flags = NONE;/* program options */
static char *g_directory = NULL; /* output directory */
static FILE *g_file = NULL;     /* input file */

/* Array of days of the week for translating a date */
static char* day_of_week[] = { "Sun", "Mon", "Tue",
                               "Wed", "Thu", "Fri", "Sat" }; 

/* Format Strings for dumping unsigned integers */
#if (SIZEOF_INT == 4)
#define SHORT_INT_FMT "%hu"
#define LONG_INT_FMT  "%u"
#else
#define SHORT_INT_FMT "%u"
#define LONG_INT_FMT  "%ul"
#endif /* SIZEOF_INT == 4 */

/* macros for dealing with program flags */
#define DEBUG_ON ((g_flags)&DBG_OUT)
#define VERBOSE_ON ((g_flags)&VERBOSE)
#define LIST_ONLY ((g_flags)&LIST)
#define USE_PATHS ((g_flags)&PATHS)
#define INTERACTIVE ((g_flags)&CONFIRM)
#define OVERWRITE_FILES ((g_flags)&OVERWRITE)
#define NUMBER_FILES ((g_flags)&NUMBERED)

/* Attr -- storing a structure, formated according to file specification */
typedef struct 
{
    uint16 type;
    uint16 name;
    size_t len;
    char* buf;
} Attr;


/* ********************
   UTILITY FUNCTIONS
   ******************** */
/* ask user for confirmation of the action */
int
confirm_action (const char *prompt, ...)
{
    if (INTERACTIVE)
    {
        int confirmed = 0;
        char buf[BUFSIZ + 1];
        va_list args;
        va_start (args, prompt);
        
#if HAVE_VPRINTF
        vfprintf (stdout, prompt, args);
#else
#  if HAVE_DOPRNT
        _doprnt (prompt, args, stdout);
#  endif /* HAVE_DOPRNT */
#endif /* HAVE_VPRINTF */
        fgets (buf, BUFSIZ, stdin);
        if (buf[0] == 'y' || buf[0] == 'Y') confirmed = 1;
        
        va_end (args);
        
        return confirmed;
    }
    return 1;
}

/* print message only when debug on */
void
debug_print (const char *fmt, ...)
{
    if (DEBUG_ON)
    {
        va_list args;
        va_start (args, fmt);
#if HAVE_VPRINTF
        vfprintf (stdout, fmt, args);
#else 
#  if HAVE_DOPRNT
        _doprnt (fmt, args, stdout);
#  endif /* HAVE_DOPRNT */
#endif /* HAVE_VPRINTF */
        va_end (args);
    }
}

/* finds a filename fname.N where N >= 1 and is not the name of an existing
   filename.  Assumes that fname does not already have such an extension */ 
char *
find_free_number (const char *fname)
{
    char *tmp = MALLOC (strlen (fname) 
                        + 1 /* '.' */
                        + 5 /* big enough for our purposes (i hope) */
                        + 1 /* NULL */);
    int counter = 1;
    struct stat buf;

    do
    {
        sprintf (tmp, "%s.%d", fname, counter++);
    }
    while (stat (tmp, &buf) == 0);
    return tmp;
}

/* write the data to a file... check for need to confirm or overwrite */
static void
write_file (const char *fname,
            const struct date *fdate,
            const char *data,
            size_t len)
{
    FILE *fp = NULL;
    char *base_fname = basename (fname);

    debug_print ("%sWRITING %s\n", 
                 ((LIST_ONLY==0)?"":"NOT "), fname);

    if (!LIST_ONLY)
    {
        if (!confirm_action ("extract %s?", base_fname)) return;
        if (!OVERWRITE_FILES)
        {
            struct stat buf;
            if (stat (fname, &buf) == 0)
            {
                if (!NUMBER_FILES)
                {
                    fprintf (stderr, 
                             "tnef: %s: Could not create file: File exists\n", 
                             fname);
                    return;
                }
                else
                {
                    char *tmp = find_free_number (fname);
                    debug_print ("Renaming %s to %s", fname, tmp);
                    write_file (tmp, fdate, data, len);
                    return;
                }
            }
        }
        
        fp = fopen (fname, "wb");
        if (fp == NULL) 
        {
            perror (fname);
            exit (1);
        }
        if (fwrite (data, 1, len, fp) != len)
        {
            perror (fname);
            exit (1);
        }
        fclose (fp);
    }
    
    if (LIST_ONLY || VERBOSE_ON)
    {
        if (LIST_ONLY && VERBOSE_ON)
        {
            /* FIXME: print out date and stuff */
            fprintf (stdout, "%11lu %04u-%02u-%02u %02u:%02u %s\n", 
                     (unsigned long)len,
                     fdate->year, fdate->month, fdate->day,
                     fdate->hour, fdate->min,
                     base_fname);
        }
        else
        {
            fprintf (stdout, "%s\n", base_fname);
        }
    }
}

static void
rename_file (const char *from,
             const char *to)
{
    debug_print ("%sRENAMEING %s to %s\n", 
                 ((LIST_ONLY==0)?"":"NOT "), from, to);

    if (!LIST_ONLY)
    {
        if (!confirm_action ("rename %s to %s?", from, to)) return;
        if (!OVERWRITE_FILES)
        {
            struct stat buf;
            if (stat (to, &buf) == 0)
            {
                fprintf (stderr, 
                         "tnef: %s: Could not rename file %s: File exists", 
                         to, from);
                return;
            }
        }
        if(rename(from, to) == -1)
        {
            perror(from);
        }
    }

    if (LIST_ONLY || VERBOSE_ON)
    {
        char *from_file = basename (from);
        char *to_file = basename (to);
        fprintf (stdout, "%s --> %s\n", from_file, to_file);
    }
}


/* 
   geti32, geti16,
   Get 16 or 32 bits from the file 
*/
static uint32
geti32 ()
{
    unsigned char buf[4];

    if (fread (buf, 4, 1, g_file) != 1) 
    {
        perror ("Unexpected end of input");
        exit (1);
    }
    return (uint32)GETINT32(buf);
}
static uint16
geti16 ()
{
    unsigned char buf[2];

    if (fread (buf, 2, 1, g_file) != 1) 
    {
        perror ("Unexpected end of input");
        exit (1);
    }
    return (uint16)GETINT16(buf);
}

/* Copy the date data from the attribute into a struct date */
static void
copy_date_from_attr (Attr* attr, struct date* dt)
{
    assert (attr);
    assert (dt);
    assert (attr->type == szDATE);

    if (attr->len >= 14)
    {
        memmove (dt, attr->buf, attr->len);
        dt->year = GETINT16((unsigned char*)&dt->year);
	    dt->month = GETINT16((unsigned char*)&dt->month);
	    dt->day = GETINT16((unsigned char*)&dt->day);
	    dt->hour = GETINT16((unsigned char*)&dt->hour);
	    dt->min = GETINT16((unsigned char*)&dt->min);
	    dt->sec = GETINT16((unsigned char*)&dt->sec);
	    dt->dow = GETINT16((unsigned char*)&dt->dow);
	} 
    else 
    {
	    fprintf (stderr, "date attribute in %s failed sanity check\n",
                 get_attr_name_str (attr->name));
	    memset (dt, 0, sizeof(*dt));
	}
}

/* dump_attr
   print attr to stderr.  Assumes that the Debug flag has been set and
   already checked */
static void 
dump_attr (Attr* attr)
{
    char *name = get_attr_name_str (attr->name);
    char *type = get_attr_type_str (attr->type);
    struct date dt;
    uint16 s;
    uint32 l;
    unsigned long i;

    fprintf (stdout, "%s [type: %s] =", name, type);

    switch (attr->type)
    {
    case szBYTE:
        for (i=0; i < attr->len; i+=1) 
        {
            if (i< 10) fprintf (stdout, " %02x", (uint8)attr->buf[i]);
            else if (i==10) fprintf (stdout, "...");
        }
        break;

    case szSHORT:
        if (attr->len < sizeof(uint16))
        {
            fprintf (stdout, "Not enough data for szSHORT");
            abort();
        }
		s = GETINT16(attr->buf);
        fprintf (stdout, " " SHORT_INT_FMT, s);
        if (attr->len > sizeof(uint16))
        {
            fprintf (stdout, " [extra data:");
            for (i = sizeof(uint16); i < attr->len; i++)
            {
                fprintf (stdout, " %02x", (uint8)attr->buf[i]);
            }
            fprintf (stdout, " ]");
        }
        break;

    case szLONG:
        if (attr->len < sizeof(uint32))
        {
            fprintf (stdout, "Not enough data for szLONG");
            abort();
        }
		l = GETINT32(attr->buf);
        fprintf (stdout, " " LONG_INT_FMT, l);
        if (attr->len > sizeof(uint32))
        {
            fprintf (stdout, " [extra data:");
            for (i = sizeof(uint32); i < attr->len; i++)
            {
                fprintf (stdout, " %02x", (uint8)attr->buf[i]);
            }
            fprintf (stdout, " ]");
        }
        break;
        

    case szWORD:
        for (i=0; i < attr->len; i+=2) 
        {
            if (i < 6) fprintf (stdout, " %04x", GETINT16(attr->buf));
            else if (i==6) fprintf (stdout, "...");
        }
        break;

    case szDWORD:
        for (i=0; i < attr->len; i+=4) 
        {
            if (i < 4) fprintf (stdout, " %08x", GETINT32(attr->buf));
            else if (i==4) fprintf (stdout, "...");
        }
        break;

    case szDATE:
        copy_date_from_attr (attr, &dt);
        fprintf (stdout, " %s %04d/%02d/%02d %02d:%02d:%02d",
                 day_of_week[dt.dow],
                 dt.year, dt.month, dt.day,
                 dt.hour, dt.min, dt.sec);
        break;

    case szTEXT:
    case szSTRING:
        {
            char* buf = MALLOC ((attr->len + 1) * sizeof (char));
            strncpy (buf, (char*)attr->buf, attr->len);
            buf[attr->len] = '\0';
            fprintf (stdout, "'%s'", buf);
            FREE (buf);
        }
        break;

    default:
        fprintf (stdout, "<unknown type> len=%lu", 
                 (unsigned long)attr->len);
    }
    fprintf (stdout, "\n");
}


/* Validate the checksum against attr.  The checksum is the sum of all the
   bytes in the attribute data modulo 65536 */
static int
check_checksum (Attr* attr, uint16 checksum)
{
    uint32 i;
    uint16 sum = 0;

    for (i = 0; i < attr->len; i++)
    {
        sum += (uint8)attr->buf[i];
    }
    sum %= 65536;
    return (sum == checksum);
}


/* takes int d and decodes the attribute from it, returns the attr, this
   functions reads from the file */
static Attr*
decode_attribute (unsigned int d) 
{
    Attr* attr = (Attr*)calloc (1, sizeof (Attr));
    size_t bytes_read = 0;
    uint16 checksum;

    assert (attr);
    if (!attr) abort();

    attr->type = (d >> 16);
    attr->name = ((d << 16) >> 16);
    attr->len = geti32();
    attr->buf = MALLOC (attr->len);

    bytes_read = fread (attr->buf, 1, attr->len, g_file);
    if (bytes_read != attr->len)
    {
        perror ("Unexpected end of input!");
        exit (1);
    }
    
    checksum = geti16();
    if (!check_checksum(attr, checksum))
    {
        fprintf (stderr, 
                 "Invalid checksum, input file may be corrupted\n");
        exit (1);
    }
    
    if (DEBUG_ON) dump_attr (attr);

    return attr;
}

/* Decodes data of type message.  This data is usally non-file related. */

static void
decode_message (void) 
{  
    uint32 d;
    Attr *attr = NULL;
    d = geti32();
    attr = decode_attribute(d);
    FREE (attr);
}



/* Decodes data of type attachment.  This is most often the actual data for
 the files in the TNEF attachment */
static void
decode_attachment (void) 
{  
    char *buf = NULL;
    uint32 d;
    static char *fname = NULL;
    static char *prevtmpfname = NULL;
    Attr *attr = NULL;
    static struct date fdate = { 0, 0, 0, 0, 0, 0, 0 };
    static unsigned long file_counter = 1;
    
    d = geti32();
    attr = decode_attribute (d);
    if (!attr) abort();
    
    switch (attr->name)
    {
    case attATTACHMODIFYDATE:
        copy_date_from_attr (attr, &fdate);
        break;

    case attATTACHTITLE:
        /* check to see if we have a filename... if not we will make a
           default and pretend that we actually got it */
        if (attr->len && *attr->buf == '\0')
        {
            debug_print ("No file name specified, using default.\n");
            attr->len = 32;     /* seems like a reasonable size */
            attr->buf = MALLOC ((attr->len + 1) * sizeof (char));
            sprintf (attr->buf, "tnef-tmp.%ld", file_counter++);
        }
        
        /* Define TEST_USE_PATHS to test the --use-path feature.  When
           defined it will pretend that all paths in the tnef file begin
           with /tmp */
#ifdef TEST_USE_PATHS
        {
            char *tmp = attr->buf;
            attr->buf = concat_fname ("/tmp", tmp);
            debug_print ("TEST_USE_PATHS: Changing file path: "
                         "'%s' -> '%s'\n",
                         tmp, attr->buf);
            FREE (tmp);
        }
#endif /* TEST_USE_PATHS */
        if (USE_PATHS)
        {
            buf = strdup (attr->buf);
        }
        else
        {
            buf = strdup(basename(attr->buf));
            if (strcmp (buf, attr->buf) != 0)
            {
                debug_print ("!!Filename contanins path: '%s'!!\n",
                             attr->buf);
            }
        }
        FREE(fname);
        fname = concat_fname (g_directory, buf);
        FREE (buf);

        /* here is where we rename the file if the filename came after the
           data */ 
        if (prevtmpfname)
        {
            rename_file (prevtmpfname, fname);
            FREE(fname);
            FREE(prevtmpfname);
        }
        break;

    case attATTACHDATA:
        if (fname == NULL)
        {
            buf=malloc((32+1) * sizeof (char));
        	sprintf (buf, "tnef-tmp.%ld", file_counter++);
        	fname = concat_fname (g_directory, buf);
            FREE(buf);

            /* since the fname was not available, save the tmp filename so
               we can rename the prevtmpfname to the real filename if we
               receive it next in attATTACHTITLE */
            prevtmpfname = strdup (fname);
        }

        write_file (fname, &fdate, attr->buf, attr->len);

        FREE(fname);
        break;
  
    default:
        break;
    }
    FREE(attr);
}


/* The entry point into this module.  This parses an entire TNEF file. */
int
parse_file (FILE* input_file, char* directory, int flags)
{
    uint32 d;
    uint16 key;

    /* store the program options in our file global variables */
    g_file = input_file;
    g_directory = directory;
    g_flags = flags;

    /* check that this is in fact a TNEF file */
    d = geti32();
    if (d != TNEF_SIGNATURE) 
    {
        fprintf (stdout, "Seems not to be a TNEF file\n");
        return 1;
    }

    /* Get the key */
    key = geti16();
    debug_print ("TNEF Key: %hx\n", key);

    /* The rest of the file is a series of 'messages' and 'attachments' */
    for (;;) 
    {
        char buf[1];
        uint8 c;

        if (fread (buf, 1, 1, g_file) == 0) break;
        c = (uint8)buf[0];

        switch (c) 
        {
        case LVL_MESSAGE:
            decode_message();
            break;
        case LVL_ATTACHMENT:
            decode_attachment();
            break;
        default:
            fprintf (stdout, "Coding Error in TNEF file\n");
            return 1;
        }
    }
    return 0;
}
